import numpy as np
import casadi as ca

def newton_method(F, z0, tol=1e-12, max_iter=100):
    """
    Newton's method for solving F(z) = 0

    F is a function with n inputs and n outputs
    z0 is the initial guess, such that z0.shape = (n,)
    tol is the convergence tolerance (infinity norm of F(z))
    max_iter is the maximum number of iterations
    """

    n = z0.shape[0]
    z = ca.SX.sym('z', n)           # symbolic variable for input z
    z_iters = [z0]                  # save all iterations of z in this list

    # compute Jacobian of F, wrap into Function
    J = ca.Function('J',[z], [ca.jacobian(F(z), z)])

    # iteration loop
    for k in range(max_iter):
        z_prev = z_iters[-1]
        
        # compute current value of F and check for convergence
        F_prev = F(z_prev).full()
        if np.linalg.norm(F_prev, np.infty) < tol:
            # we have found a solution
            break

        # compute next guess by solving linearized system
        z_next = z_prev - ca.solve(J(z_prev), F_prev)
        z_iters.append(z_next.full())

    if k == max_iter - 1:
        print('max iter reached')
    else:
        print('converged in {} iterations'.format(k))

    return z_iters
